use std::net::SocketAddr;
use std::path::PathBuf;
use std::sync::Arc;

use axum::{
    extract::{Query, State},
    http::{header::CONTENT_TYPE, StatusCode},
    response::{AppendHeaders, IntoResponse},
    routing::get,
    Json, Router,
};

use bloomfilter::Bloom;

use color_eyre::eyre::Result;

use clap::Parser;

use serde::Deserialize;

use tower_http::catch_panic::CatchPanicLayer;
use tower_http::trace::TraceLayer;

use tracing_subscriber::filter::{EnvFilter, LevelFilter};
use tracing_subscriber::fmt::format::FmtSpan;
use tracing_subscriber::prelude::*;

use tracing::info;

mod bloom;

#[derive(Parser)]
struct CliConfig {
    #[clap(long, env)]
    bloom_filter_path: PathBuf,

    #[clap(long, env)]
    bind_address: Option<SocketAddr>,
}

struct AppState {
    bloom_filter: Bloom<String>
}

fn setup_registry() {
    let envfilter = EnvFilter::builder()
        .with_default_directive(LevelFilter::DEBUG.into())
        .from_env_lossy();
    tracing_subscriber::registry()
        .with(envfilter)
        .with(tracing_subscriber::fmt::layer().with_span_events(FmtSpan::CLOSE))
        .with(tracing_error::ErrorLayer::default())
        .init();
}

async fn index() -> impl IntoResponse {
    (
        AppendHeaders([(CONTENT_TYPE, "text/html")]),
        include_str!("index.html"),
    )
}

#[tracing::instrument(skip(state))]
async fn check_hash(hash: &str, state: &AppState) -> bool {
    let bloom_filter = &state.bloom_filter;
    bloom_filter.check(&hash.to_string())
}

#[derive(Deserialize)]
struct CheckQuery {
    sha256: String,
}

// Note: Exposes *zero* information of potential errors to clients.
#[tracing::instrument(skip(query, state))]
async fn check_hash_slug(
    query: Query<CheckQuery>,
    State(state): State<Arc<AppState>>,
) -> (StatusCode, Json<bool>) {
    let result = check_hash(&query.sha256, &state).await;
    let status_code = match state.bloom_filter.check(&query.sha256.to_string()) {
        true => StatusCode::OK,
        false => StatusCode::NOT_FOUND,
    };
    (status_code, Json(result))
}

#[tokio::main]
async fn main() -> Result<()> {
    setup_registry();
    color_eyre::install()?;

    let cli_config = CliConfig::parse();
    let addr = cli_config
        .bind_address
        .unwrap_or_else(|| (std::net::Ipv4Addr::new(127, 0, 0, 1), 8000).into());

    info!("loading bloom filter");

    let state = AppState {
        bloom_filter: bloom::load(cli_config.bloom_filter_path.as_ref())?
    };

    let app = Router::new()
        .route("/", get(index))
        .route("/check", get(check_hash_slug))
        .with_state(Arc::new(state))
        .layer(CatchPanicLayer::new())
        .layer(TraceLayer::new_for_http());

    info!("server go nyoom");

    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await?;

    Ok(())
}
